static void lonusb_dl_formware(struct usb_device *usbdev)
{
	const struct firmware *fw;
	const char *fname;
	int ret;

	switch (le16_to_cpu(usbdev->descriptor.idProduct)) 
    {
	case GESYTEC_LPU_PID:
		fname = FIRMWARE_FX;
		break;
	case GESYTEC_LPM2_PID:
		fname = FIRMWARE_FX2;
		break;
	default:
		dev_err(&usbdev->dev, "Missing FW for id: %u\n", le16_to_cpu(usbdev->descriptor.idProduct));
		return;
	};

	ret = request_firmware(&fw, fname, &usbdev->dev);
	if (ret) 
    {
		dev_err(&usbdev->dev, "Can't load firmware %s, error: %d\n", fname, ret);
		return;
	}

	ret = ezusb_load_firmware(usbdev, fw->data);
	release_firmware(fw);
	if (ret)
		dev_err(&usbdev->dev, "Firmware download failed %d\n", ret);
}

static int CreateIntUrb(PDEVICE_EXTENSION dev)
{
	int pipe;

	dev->read_urb = usb_alloc_urb(0, GFP_KERNEL);
	if (!dev->read_urb) 
    {
		pr_err("No free (read) urbs available");
		return -ENOMEM;
	}
	pipe = usb_rcvintpipe(dev->usbdev, dev->intr_in_ep_addr);
	dev->bulk_in_size = usb_maxpacket(dev->usbdev, pipe, usb_pipeout(pipe));

	dev->intr_in_buff = kmalloc(dev->bulk_in_size, GFP_KERNEL);
	if (!dev->intr_in_buff) 
    {
		pr_err("Couldn't allocate intr_in_buff");
		return -ENOMEM;
	}
	usb_fill_int_urb(dev->read_urb, dev->usbdev, pipe,
			dev->intr_in_buff, dev->bulk_in_size,
			lonusb_read_int_callback, dev,
			dev->bInterval);
	return 0;
}

static int lonusb_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
	struct usb_device *usbdev = interface_to_usbdev(interface);
	struct usb_host_interface *if_desc;
	PDEVICE_EXTENSION dev = NULL;
	struct usb_endpoint_descriptor *endpoint;
	int pipe;
	int i;
	int ret;

	if (usbdev->descriptor.iProduct == 0) 
    {
		lonusb_dl_formware(usbdev);
		ret = -ENODEV;
		goto exit;
	}

	dev = kzalloc(sizeof(DEVICE_EXTENSION), GFP_KERNEL);
	if (dev == NULL) 
    {
		dev_err(&usbdev->dev, "Out of memory");
		ret = -ENOMEM;
		goto exit;
	}

    dev->dev = usbdev->dev;
	mutex_init(&dev->mutex);
	dev->usbdev = usbdev;
	dev->interface = interface;
	dev->remove_pending = false;

    qinit(&dev->ReadBufferQueue, maxqueuecount);
	init_waitqueue_head(&dev->rd_wait);
	init_waitqueue_head(&dev->wr_wait);

	/* set up the endpoint information */
	if_desc = interface->cur_altsetting;
	for (i = 0; i < if_desc->desc.bNumEndpoints; ++i) 
    {
		endpoint = &if_desc->endpoint[i].desc;
		if (usb_endpoint_is_int_in(endpoint) && !dev->intr_in_buff) 
        {
			/* we found a interrupt in endpoint */
			dev->intr_in_ep_addr = endpoint->bEndpointAddress;
			dev->bInterval = endpoint->bInterval;
			ret = CreateIntUrb(dev);
			if (ret)
				goto error;
		}

		if (usb_endpoint_is_bulk_out(endpoint) && !dev->bulk_out_buffer) 
            {
			/* we found a bulk out endpoint */
#if REUSE_URB
            dev->write_urb = usb_alloc_urb(0, GFP_KERNEL);
            if (!dev->write_urb)
                {
                pr_err("No free (write) urbs available");
                ret = -ENOMEM;
                goto error;
                }
#endif
			pipe = usb_sndbulkpipe(usbdev, endpoint->bEndpointAddress);
			dev->bulk_out_size = usb_maxpacket(dev->usbdev, pipe, usb_pipeout(pipe));
			dev->bulk_out_size = USB_WRITEBUFFER_SIZE;  // use maximum needed, because usb_submit_urb() does internal divide it

			dev->bulk_out_endpointAddr = endpoint->bEndpointAddress;
			dev->bulk_out_buffer = kmalloc(dev->bulk_out_size, GFP_KERNEL);
			if (!dev->bulk_out_buffer) 
            {
				pr_err("Couldn't allocate bulk_out_buffer");
				ret = -ENOMEM;
				goto error;
			}
		}
	}

	/* we can register the device now, as it is ready */
	usb_set_intfdata(interface, dev);

	/* we can register the device now, as it is ready */
	ret = usb_register_dev(interface, &lonusb_class);
	if (ret) 
    {
		/* something prevented us from registering this driver */
		pr_err("Not able to get a minor for this device.");
		usb_set_intfdata(interface, NULL);
		goto error;
	}

	ret = 0;
#if USE_PROC
	{
  #ifdef CONFIG_USB_DYNAMIC_MINORS
        // see minor_base at http://www.makelinux.net/ldd3/chp-13-sect-4  
        int minor = interface->minor;
  #else
        int minor = interface->minor - lonusb_class.minor_base;
  #endif
		sprintf(dev->DevStats.ProcDevName, PROC_DEV_NAME "%d" /*lonusb_class.name*/, minor);
		if (HasEasylonProcDir)
        {
            //printk(KERN_DEBUG "proc_create_data(%s, %p):\n", dev->DevStats.ProcDevName, dev);
			proc_create_data(dev->DevStats.ProcDevName, 0, NULL, &lonusb_proc_fops_dev, dev);
        }
	}

	ProcStats.probe_cnt++ ;
#endif
 	goto exit;

error:
	lonusb_delete(dev);
	dev = NULL;
exit:
	return ret;
}

static void lonusb_disconnect(struct usb_interface *interface)
{
	PDEVICE_EXTENSION dev = usb_get_intfdata(interface);
	usb_set_intfdata(interface, NULL);

	/* give back our minor */
	usb_deregister_dev(interface, &lonusb_class);

	mutex_lock(&dev->mutex);

	dev->remove_pending = true;
	wake_up_interruptible(&dev->rd_wait);
	wake_up_interruptible(&dev->wr_wait);

	/* if the device is not opened, then we clean up right now */
	if (!dev->open_count) 
    {
		mutex_unlock(&dev->mutex);
		lonusb_delete(dev);
	} else 
    {
		/* shutdown any bulk writes that might be going on */
		usb_kill_urb(dev->read_urb);
#if REUSE_URB
        usb_kill_urb (dev->write_urb);
#endif
		dev->usbdev = NULL;
		mutex_unlock(&dev->mutex);
	}
#if USE_PROC
	if (HasEasylonProcDir)
		remove_proc_entry(dev->DevStats.ProcDevName, NULL);

	ProcStats.disconnect_cnt++ ;
#endif
}

static int __init lonusb_init(void)
{
#if USE_PROC
    if (CreateEasylonProcDir())
        proc_create_data(PROC_DEV_NAME, 0, NULL, &lonusb_proc_fops_all, NULL); 
#endif

 	return usb_register(&lonusb_driver);
}

static void __exit lonusb_cleanup(void)
{
#if USE_PROC
	if (HasEasylonProcDir)
		remove_proc_entry(PROC_DEV_NAME, NULL);
    RemoveEasylonProcDir();
#endif
    usb_deregister(&lonusb_driver);
}
module_exit(lonusb_cleanup);
module_init(lonusb_init);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
