/*
 * Linux driver for Easylon USB Interfaces
 *
 * Copyright (c)2003 Gesytec GmbH, Volker Schober (info@gesytec.de)
 *
 */

#include "lonusb.h"

#include "dynbuf.c"

#include "driverdata.c"
#include "proc.c"
#include "ezusb.c"
#include "driverentry.c"
#include "readwrite.c"


static void lonusb_debug_data(const char *function, int size, const unsigned char *data)
{
	if (!debug)  return;
	//printk(KERN_DEBUG "%s ", function);
	print_hex_dump(KERN_DEBUG, function, DUMP_PREFIX_OFFSET, 32, 1, data, size, false);
}

static void lonusb_delete(PDEVICE_EXTENSION dev)
{
	kfree(dev->intr_in_buff);
	kfree(dev->bulk_out_buffer);
#if REUSE_URB
    if (dev->write_urb != NULL)
        {
		dev_dbg(&dev->dev, "usb_free_urb(dev->write_urb)");
        usb_free_urb (dev->write_urb);
        }
#endif
	if (dev->read_urb != NULL) 
    {
		dev_dbg(&dev->dev, "usb_free_urb(dev->read_urb)");
		usb_free_urb(dev->read_urb);
	}

	dev->RemainDataSize = 0;
    qexit(&dev->ReadBufferQueue);
    kfree(dev);
}

static int lonusb_open(struct inode *inode, struct file *file)
{
	int retval = 0;
	PDEVICE_EXTENSION dev = NULL;
	int subminor = iminor(inode);
	struct usb_interface *interface = usb_find_interface(&lonusb_driver, subminor);

	if (!interface) 
    {
		dev_err(&dev->dev, "error, can't find device for minor %d", subminor);
		retval = -ENODEV;
		goto exit;
	}

	dev = usb_get_intfdata(interface);
	if (!dev) 
    {
		retval = -ENODEV;
		goto exit;
	}

    {
        dev_dbg(&dev->dev, "lonusb_open(subminor=%d) minor_base=%d\n", subminor, lonusb_class.minor_base);
    }

	dev->minor = subminor;

	mutex_lock(&dev->mutex);

	if ((++dev->open_count) > 1) 
    {
		retval = -EBUSY;
		goto error;
	}

	file->private_data = dev;

	dev->RemainDataSize = 0;

	retval = usb_submit_urb(dev->read_urb, GFP_KERNEL);
    dev->DevStats.open++;

error:
	if (retval < 0)
		--dev->open_count;

	mutex_unlock(&dev->mutex);
exit:
	return retval;
}

static int lonusb_release(struct inode *inode, struct file *file)
{
	PDEVICE_EXTENSION dev = file->private_data;
	int retval = 0;

	if (dev == NULL)
		return -ENODEV;

    dev_dbg(&dev->dev, "lonusb_release(subminor=%d)\n", dev->minor);

    mutex_lock(&dev->mutex);
	if (dev->usbdev == NULL) 
    {
		/* the device was unplugged before the file was released */
		mutex_unlock(&dev->mutex);
		lonusb_delete(dev);
		return 0;
	}

	if (--dev->open_count <= 0) 
    {
		/* shutdown any bulk writes that might be going on */
		usb_kill_urb(dev->read_urb);

#if REUSE_URB
        usb_kill_urb (dev->write_urb);
#endif
		dev->RemainDataSize = 0;
        qflush(&dev->ReadBufferQueue);
		dev->open_count = 0;
	}

	mutex_unlock(&dev->mutex);
    dev->DevStats.close++;
	return retval;
}
