static unsigned lonusb_poll(struct file *file, struct poll_table_struct *wait)
{
	PDEVICE_EXTENSION dev = file->private_data;
	unsigned mask = 0;

    dev_dbg(&dev->dev, "lonusb_poll()\n");
	dev->DevStats.poll++;

    /* verify that the device wasn't unplugged */
	if (dev->usbdev == NULL || dev->remove_pending)
	{
		dev->DevStats.poll_error++;
		return POLLERR | POLLHUP;
	}

	if (file->f_mode & FMODE_READ)
	{
		dev->DevStats.poll_read++;
        poll_wait(file, &dev->rd_wait, wait);
    }
	if (file->f_mode & FMODE_WRITE)
	{
		dev->DevStats.poll_write++;
        poll_wait(file, &dev->wr_wait, wait);
    }

	/* verify that the device wasn't unplugged */
	if (dev->usbdev == NULL || dev->remove_pending)
	{
		dev->DevStats.poll_error++;
		return POLLERR | POLLHUP;
	}

    if (qhasdata(&dev->ReadBufferQueue))
	{
		dev->DevStats.poll_pollin++;
		mask |= POLLIN | POLLRDNORM;
	}
#if REUSE_URB
    if (dev->write_urb->status != -EINPROGRESS)
#else
	if (!atomic_read(&dev->write_count))
#endif
	{
		dev->DevStats.poll_pollout++;
		mask |= POLLOUT | POLLWRNORM;
	}

	return mask;
}

static void lonusb_read_int_callback(struct urb *urb)
{
	PDEVICE_EXTENSION dev = urb->context;
	struct usb_device *usbdev = dev->usbdev;
    NTSTATUS status;
    unsigned int TransferLen;
	int ret;

    switch (urb->status)
    {
    case 0:
        /* success */
        break;
    case -ECONNRESET:
    case -ENOENT:
    case -ESHUTDOWN:
        /* this urb is terminated, clean up */
        dev->RemainDataSize = 0;
        return;
    default:
        dev->RemainDataSize = 0;
        goto exit;
    }

	TransferLen = urb->actual_length;
	TransferLen += dev->RemainDataSize;

	if (TransferLen == 0)
		goto exit;

	if (TransferLen > sizeof(dev->intdata))
    {
		dev_dbg(&usbdev->dev, "TransferLen=%u > sizeof(dev->intdata)=%zu", TransferLen, sizeof(dev->intdata));
		TransferLen = dev->RemainDataSize = 0;
		goto exit;
	}
	memcpy(&dev->intdata[dev->RemainDataSize], dev->intr_in_buff, urb->actual_length);
	if (urb->actual_length == dev->bulk_in_size)
    {
		dev->RemainDataSize = TransferLen;
		dev_dbg(&usbdev->dev, " RemainDataSize=%u:", dev->RemainDataSize);
		goto exit;
	}
	dev->RemainDataSize = 0;

    lonusb_debug_data ("<< ", TransferLen, dev->intdata);
    dev->DevStats.usb_in++;
    status = qwrite(&dev->ReadBufferQueue, dev->intdata, TransferLen, FALSE);
    wake_up_interruptible(&dev->rd_wait);
    dev_dbg(&usbdev->dev, "wake_up_interruptible\n");
exit:
	ret = usb_submit_urb(urb, GFP_ATOMIC);
	if (ret)
		dev_err(&usbdev->dev, "usb_submit_urb()=%d", ret);
}


static ssize_t lonusb_read(struct file *file, char __user *buffer, size_t count, loff_t *ppos)
{
	PDEVICE_EXTENSION dev = file->private_data;
	int ret = 0;

    dev_dbg(&dev->dev, "lonusb_read():\n");
	dev->DevStats.read_in++;

	mutex_lock(&dev->mutex);
	if (dev->usbdev == NULL || dev->remove_pending)
    {
		mutex_unlock(&dev->mutex);
        dev->DevStats.read_error++;
		return -ENODEV;
	}

    if (qempty(&dev->ReadBufferQueue))
  {
		mutex_unlock(&dev->mutex);
		if (file->f_flags & O_NONBLOCK)
        {
            dev->DevStats.read_error++;
			return -EAGAIN;
        }

        ret = wait_event_interruptible(dev->rd_wait, (qhasdata(&dev->ReadBufferQueue) || dev->remove_pending));
        dev_dbg(&dev->dev, "lonusb_read(): wait_event_interruptible()=%d\n", ret);
		if (ret < 0)
        {
            dev->DevStats.read_error++;
			return ret;
        }
		mutex_lock(&dev->mutex);
		if (dev->usbdev == NULL || dev->remove_pending)
        {
			mutex_unlock(&dev->mutex);
            dev->DevStats.read_error++;
			return -ENODEV;
		}
	}

    ret = qread(&dev->ReadBufferQueue, buffer, count, TRUE);
	mutex_unlock(&dev->mutex);
    dev_dbg(&dev->dev, "lonusb_read()=%d\n", ret);
    dev->DevStats.read_ok++;
	return ret;
}

static void lonusb_write_bulk_callback(struct urb *write_urb)
{
	PDEVICE_EXTENSION dev = write_urb->context;

	lonusb_debug_data(">> ", write_urb->transfer_buffer_length, write_urb->transfer_buffer);
#if !REUSE_URB
	atomic_dec(&dev->write_count);
#endif
	wake_up_interruptible(&dev->wr_wait);

	if (write_urb->status != 0)  /* USB_ST_NOERROR */
    {
        dev->DevStats.write_error++;
		dev_dbg(&dev->dev, "write bulk error: %d", write_urb->status);
    }
    else
        dev->DevStats.write_ok++;
	return;
}

static ssize_t lonusb_write(struct file *file, const char __user *buffer, size_t count, loff_t *ppos)
{
	PDEVICE_EXTENSION dev = file->private_data;
	struct usb_device *usbdev;
	struct urb *write_urb = NULL;
	ssize_t bytes_written = 0;
	int retval = 0;
	int offset = 0;

    dev->DevStats.write_in++;
	mutex_lock(&dev->mutex);
	usbdev = dev->usbdev;
	if (!usbdev || dev->remove_pending)
    {
		retval = -ENODEV;
		goto error_up;
	}

	if (count == 0)
		goto error_up;

	offset = 0;
	while (offset < count)
    {
		/* see if we are already in the middle of a write */
#if REUSE_URB
        if (dev->write_urb->status == -EINPROGRESS)
#else
		if (atomic_read(&dev->write_count))
#endif
        {
			mutex_unlock(&dev->mutex);
			if (file->f_flags & O_NONBLOCK)
				return -EAGAIN;

#if REUSE_URB
            if (wait_event_interruptible(dev->wr_wait, dev->write_urb->status != -EINPROGRESS))
#else
			if (wait_event_interruptible(dev->wr_wait, !atomic_read(&dev->write_count)))
#endif
            {
				return -ERESTARTSYS;
			}
			mutex_lock(&dev->mutex);
		}
#if REUSE_URB
        write_urb = dev->write_urb;
#else
		atomic_inc(&dev->write_count);
		write_urb = usb_alloc_urb(0, GFP_KERNEL);
		if (!write_urb)
        {
			dev_err(&usbdev->dev, "No free (write) urbs available");
			retval = -ENOMEM;
			goto error_up;
		}
#endif
		bytes_written = ((count-offset) > dev->bulk_out_size) ? dev->bulk_out_size : count-offset;

		/* copy the data (from userspace into our urb */
		if (copy_from_user(dev->bulk_out_buffer, buffer + offset, bytes_written))
        {
			retval = -EFAULT;
			goto error;
		}
		offset += bytes_written;
		lonusb_debug_data(">  ", bytes_written, dev->bulk_out_buffer);

		usb_fill_bulk_urb(write_urb, usbdev,
			usb_sndbulkpipe(usbdev, dev->bulk_out_endpointAddr),
			dev->bulk_out_buffer, bytes_written,
			lonusb_write_bulk_callback, dev);

#if REUSE_URB
        write_urb->status = 0;	// needed!
#endif
		retval = usb_submit_urb(write_urb, GFP_ATOMIC);
		if (retval)
        {
			dev_err(&usbdev->dev, "failed submitting write urb, %d", retval);
		} else
        {
			retval = offset;
			goto exit;
		}
	}
error:
#if REUSE_URB
exit:
#else
	atomic_dec(&dev->write_count);
exit:
	usb_free_urb(write_urb);
#endif
error_up:
	mutex_unlock(&dev->mutex);
	return retval;
}

static long lonusb_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	int retval;

	switch (cmd)
    {
    char tmpBuf[80];
    case IOCTL_VERSION:
        retval = sprintf(tmpBuf, DRIVER_DESC " " DRIVER_VERSION ", Linux %d.%d.%d",
            (LINUX_VERSION_CODE >> 16) & 0xFF,
            (LINUX_VERSION_CODE >> 8) & 0xFF,
            LINUX_VERSION_CODE & 0xFF ) + 1;
        if (copy_to_user ((char *)arg, tmpBuf, retval))
            retval = -EFAULT;
        break;
	default:
		retval = -EINVAL;
	}
	return retval;
}

