/*********************************************************************
 *                
 * Filename:      slta-attach.c
 * Version:       
 * Description:   
 * Status:        Experimental.
 * Author:        Michael Tross <mtr@nets-solutions.de>
 * Created at:    Mon May 15 17:09:48 CEST 2006
 * Modified at:   
 * Modified by:   
 * Sources:       irattach
 *
 *     Copyright (c) 1997, 1999-2000 Dag Brattli <dagb@cs.uit.no>, 
 *     Copyright (c) 2006, Michael Tross <mtr@nets-solutions.de>
 *     All Rights Reserved.
 *     
 *     This program is free software; you can redistribute it and/or 
 *     modify it under the terms of the GNU General Public License as 
 *     published by the Free Software Foundation; either version 2 of 
 *     the License, or (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *     
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *     
 ********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <syslog.h>
#include <termios.h>
#include <unistd.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <sys/utsname.h>

#include "slta-tty.h"


#ifndef N_SLTA
#define N_SLTA N_MASC /* This one should go in .../asm/termio.h */
#endif /* N_SLTA */


/* External prototypes */
extern void fork_now(int ttyfd);
extern int set_sysctl_param(char *name, char *value);
extern int execute(char *msg, char *cmd);
/* Internal prototypes */

#define VERSION "1.00"


extern char *optarg;
extern int optind;

static int devfd = -1;		/* File descriptor for the tty device */
static int fdflags = -1;	/* Current file descriptor flags */
static int initfdflags = -1;	/* Initial file descriptor flags */
static int initdisc = -1;	/* Initial line discipline */
static struct termios termsave;	/* Saved tty termios */
static int termvalid = -1;	/* If termsave is valid */

/* Default path for pid file */
char *pidfile = "/var/run/slta/pid";

/* Serial Device Name of the device we manage */
char device[32];

/* Serial interface setup */
int baudrate = 0;


/********************* SIGNAL & ERROR HANDLING *********************/

/*
 * Put everything back in the state we found it
 */
void cleanup(int signo)
{
	switch (signo) {
	case SIGTERM:
	case SIGINT:
		syslog(LOG_INFO, "got SIGTERM or SIGINT\n");
		break;
	case SIGHUP:
		syslog(LOG_INFO, "got SIGHUP\n");
		break;
	default:
		break;
	}


	/* Reset the fd termios struct ++ */
	if (devfd != -1) {
		fdflags &= ~O_NONBLOCK;
		if (fcntl (devfd, F_SETFL, fdflags) != 0) {
			syslog (LOG_ERR, "fcntl: fdflags: %m");
		}
		if (initdisc != -1)
			if (ioctl(devfd, TIOCSETD, &initdisc) < 0){
				fprintf(stderr, "Oups ! Can't set back original link discipline...\n");
				syslog(LOG_ERR, "ioctl: set_inidisc: %m");
			}
		if (termvalid != -1)
			if (tcsetattr(devfd, TCSANOW, &termsave) != 0) {
				syslog (LOG_ERR, "tcsetattr: %m");
			}
		if (fcntl (devfd, F_SETFL, initfdflags) != 0) {
			syslog (LOG_ERR, "fcntl: initfdflags: %m");
		}
		close(devfd);
		devfd = -1;
	}

	/* Delete pid file */
	unlink(pidfile);

	syslog(LOG_INFO, "exiting ...\n");

	if (signo == -1)
		exit(1);
	else
		exit(0);
}

/* To avoid recursion, clean_exit() should no be used above this
 * point - Jean II */
static inline void clean_exit(int status)
{
	cleanup(status);
	//exit(-1); /* should not get here */
}

static void print_usage(void)
{
	fprintf(stderr, "Usage: slta-attach <dev> [-v] [-h] [-b baudrate] [-p pidfile]\n");
	fprintf(stderr, "       <dev> is tty name, device name or module name\n");
	fprintf(stderr, "\n");
}

/************************ MODULES MANAGEMENT ************************/

/*
 * Function load_module (name)
 *
 *    Tries to load (modprobe) the module with the given name
 *
 */
static int load_module(char *name)
{
	char msg[128], cmd[512];
	int ret;
	
	sprintf(msg, "Trying to load module %s", name);
	sprintf(cmd, "/sbin/modprobe %s", name);

	ret = execute(msg, cmd);

	return ret;
}

/************************** TTY MANAGEMENT **************************/

/*
 * Function establish_slta (ttyfd)
 * 
 *    Turn the serial port into a slta interface.
 */
static void establish_slta (int ttyfd) 
{
	int sltadisc = N_SLTA;

#if 0	
	if (ioctl(ttyfd, TIOCEXCL, 0) < 0) {
		syslog (LOG_WARNING, "ioctl(TIOCEXCL): %m");
	}
#endif
	if (ioctl(ttyfd, TIOCNXCL, 0) < 0) {
		syslog (LOG_WARNING, "ioctl(TIOCEXCL): %m");
	}

	if (ioctl(ttyfd, TIOCGETD, &initdisc) < 0) {
		syslog(LOG_ERR, "ioctl(TIOCGETD): %m");
		clean_exit(-1);
	}
	
	if (ioctl(ttyfd, TIOCSETD, &sltadisc) < 0){
		fprintf(stderr,  
			 "Maybe you don't have SLTA kernel module loaded?\n");
		syslog(LOG_ERR, "slta-attach: tty: set_disc(%d): %s\n", 
			sltadisc, strerror(errno));
		clean_exit(-1);
	}
}

/*
 * Function tty_configure (tios)
 *
 *    Put a slta line discipline in a transparent mode. 
 *
 */
static int tty_configure(struct termios *tios) 
{
	tios->c_cflag     = CS8|CREAD|B115200|CLOCAL;
	
	/* Ignore break condition and parity errors */
	tios->c_iflag     = IGNBRK | IGNPAR;
	tios->c_oflag     = 0;
	tios->c_lflag     = 0; /* set input mode (non-canonical, no echo,..) */
	tios->c_cc[VMIN]  = 1; /* num of chars to wait for, before delivery */
	tios->c_cc[VTIME] = 0; /* timeout before delivery */

	return(0);
}

/*
 * Function init_slta (ttyfd)
 *
 *    Initialize SLTA line discipline
 *
 */
static void init_slta_ldisc(int ttyfd) 
{
	struct termios tios;
	
	/* Get TTY configuration */
	if (tcgetattr(ttyfd, &tios) != 0) {
		syslog (LOG_ERR, "tcgetattr: %m");
		clean_exit(-1);
	}

	/* Save the original values */
	memcpy(&termsave, &tios, sizeof(struct termios));
	termvalid = 0;

	tty_configure(&tios);
	
	/* tcflush(ttyfd, TCIFLUSH); */
	if (tcsetattr(ttyfd, TCSAFLUSH, &tios) < 0) {
		syslog(LOG_ERR, "tcsetattr: %m");
		clean_exit(-1);
	}
}

/*
 * Function start_tty(ttyfd)
 *
 *    Set up the serial device to be the slta interface.
 *
 * This needs to be called after the fork.
 */
static void start_tty(int ttyfd) 
{
	int cloexec;

	/* Set new flags -> blocking mode while we configure */
	fdflags &= ~O_NONBLOCK;
	if (fcntl(ttyfd, F_SETFL, fdflags) == -1) {
		syslog(LOG_WARNING, "Couldn't set device fd flags: %m");
	}
		
	/* We need to make sure the devfd/ttyfd will not be inherited by the
	 * shells which get vfork'ed when we use popen(3) to execute commands
	 * like sysctl and modprobe on behalf of irattach. In the daemon mode
	 * we got most likely fd=0 which would be reused for stdin by /bin/sh
	 * otherwise!
	 * F_SETFD belongs to the fd, not file: no need to save and restore
	 * later.
	 * Martin
	 */
	if ((cloexec = fcntl(devfd, F_GETFD)) == -1) {
		syslog(LOG_ERR,
		       "Couldn't get device fd close-on-exec flag: %m");
		clean_exit(-1);
	}
	cloexec |= FD_CLOEXEC;
	if (fcntl(devfd, F_SETFD, cloexec) == -1) {
		syslog(LOG_ERR,
		       "Couldn't set device fd close-on-exec flag: %m");
		clean_exit(-1);
	}

	/* Set up the serial device as a slta interface */	
	init_slta_ldisc(ttyfd);
	establish_slta(ttyfd);

	/* Give it time to set up its terminal */
	sleep(1);

	/* baudrate */
	if (baudrate > 0) {
		syslog(LOG_INFO, "set baudrate to %u", baudrate);
		if (ioctl(ttyfd, SLTA_IOCBAUD, &baudrate) < 0) {
			syslog(LOG_ERR, "ERROR: couldn't set baudrate to %u", baudrate);
		}
	}

#if 0	
	/*
	 *  Set device for non-blocking reads.
	 */
	if (fcntl(ttyfd, F_SETFL, fdflags | O_NONBLOCK) == -1) {
		syslog(LOG_ERR, 
		       "Couldn't set device to non-blocking mode: %m");
		clean_exit(-1);
	}
#endif
}

/*
 * Function open_tty(ttyfd)
 *
 *    Open the desired tty, save its state.
 *
 * This needs to be called before the fork.
 */
static int open_tty(char *dev) 
{
	int ttyfd;

	/* Open the serial device */
	ttyfd = open(dev, O_NONBLOCK | O_RDWR, 0);
	if (ttyfd < 0) {
		fprintf(stderr, "Failed to open device %s: %s\n",
			dev, strerror(errno));
		exit(-1);
	}

	/* Save old flags
	 * Need to happen first, because we need to restore them when leaving
	 * via clean_exit() */
	if ((fdflags = fcntl(ttyfd, F_GETFL)) == -1) {
		fprintf(stderr, "Couldn't get device %s flags: %s\n",
			dev, strerror(errno));
		close(ttyfd);
		exit(-1);
	}
	initfdflags = fdflags;

	return(ttyfd);
}

/******************************* MAIN *******************************/

/*
 * Function main (argc, )
 *
 *    Main function
 *
 */
int main(int argc, char *argv[]) 
{
	/*struct utsname buf;*/
	int	tty = 0;	/* True if first arg is a tty */
//	int	discovery = -1;	/* True if discovery requested */
	int	c;

	//printf("%s\n", VERSION);
	if (argc < 2) {
		print_usage();
		exit(-1);
	}

	/* First arg is device name. Save it now, because in some cases
	 * getopt() will remove it... */
	strncpy(device, argv[1], sizeof(device));
	device[sizeof(device)-1] = '\0';

	/* Look for options */
	/* Do this before processing device, to handle "-h" and -v"
	 * properly. Jean II */
	while ((c = getopt(argc, argv, "b:hsvp:")) != -1) {
		switch (c) {
		case 'b':
			baudrate = atoi(optarg);
			break;
		case 'p':
			pidfile = strdup(optarg);
			break;
#if 0
		case 's':
		       	/* User wants to start discovery */
			discovery = 1;
			break;
#endif
		case 'v':
			printf("Version: %s\n", VERSION);
			exit(0);
		case 'h':
			print_usage();
			exit(0);
		default:
			print_usage();
			exit(-1);
		}
	}

	/* Check if the device is a tty */
	if(strncmp("/dev", device, 4) == 0) {
		/* We are managing a tty ! */
		tty = 1;

		/* Create tty channel and make it exist */
		devfd = open_tty(device);
	} else {
		/* Check for a irda device name */
		if((strncmp("slta", device, 4) == 0) &&
		   (isdigit(device[4]))) {
			/* May fail if no alias in /etc/modules.conf.
			 * Ignore, because the user may load module by hand.
			 * Jean II */
			load_module(device);
		}
	}

	/* Go as a background daemon */
	fork_now(devfd);

	/* --- Deamon mode --- */
	/* We can no longer print out directly to the terminal, we
	 * now must use the syslog facility.
	 * Jean II */

	/* Trap signals */
	if (signal(SIGHUP, cleanup) == SIG_ERR)
		syslog(LOG_INFO, "signal(SIGHUP): %m");
	if (signal(SIGTERM, cleanup) == SIG_ERR)
		syslog(LOG_INFO, "signal(SIGTERM): %m");
	if (signal(SIGINT, cleanup) == SIG_ERR)
		syslog(LOG_INFO, "signal(SIGINT): %m");

	/* If device is a tty */
	/* We may want to move that before the fork(), except for the sleep */
	if (tty) {
		/* Setup tty channel */
		start_tty(devfd);

#if 0
		/* Get device name corresponding to the tty */
		if (ioctl(devfd, SLTA_IOCGNAME, &info) < 0) {
			syslog(LOG_ERR, "Are you using an old kernel?");
			clean_exit(-1);
		}
		strncpy(device, info.name, 20);
#endif
	}

#if 0
	/* Use hostname as device name */
	if (uname(&buf) == 0)
		set_sysctl_param("devname", strtok(buf.nodename, "."));

	/* User may want to start discovery */
	if(discovery > 0)
		set_sysctl_param("discovery", "1");
#endif

	/*
	 *  Loop forever and wait for kill or ctrl-C since closing this 
	 *  process will also close all open files for this process
	 *  which will in turn close the tty used for SLTA which is not
	 *  really what we want :-)
	 */
	while (1)
		pause();

	return 0;
}
